@@ -1,7 +1,7 @@
 #![feature(hash_drain_filter,async_closure)]
 use std::collections::HashMap;
 use lru::LruCache;
-use async_std::{prelude::*,fs,sync::{RwLock,Arc,Mutex}};
+use async_std::{prelude::*,fs,sync::{RwLock,Arc,Mutex},io};
 use desert::{varint,ToBytes,FromBytes};
 
 mod storage;
@@ -45,24 +45,34 @@ pub struct XQ<S,R> where S: RW, R: Record {
   root: QTree,
   quad_cache: LruCache<QuadId,HashMap<RecordId,R>>,
   quad_updates: Arc<RwLock<HashMap<QuadId,Option<HashMap<RecordId,R>>>>>,
+  quad_update_age: HashMap<QuadId,usize>,
+  quad_update_count: u64,
+  quad_count: HashMap<QuadId,u64>,
   next_quad_id: QuadId,
-  // todo: quad_id_cache
   id_cache: LruCache<IdBlock,HashMap<RecordId,QuadId>>,
   id_updates: Arc<RwLock<HashMap<IdBlock,HashMap<RecordId,QuadId>>>>,
+  id_update_age: HashMap<IdBlock,usize>,
+  id_update_count: u64,
+  id_count: HashMap<IdBlock,u64>,
   missing_updates: HashMap<RecordId,R>,
   missing_count: usize,
   next_missing_id: u64,
   fields: Fields,
 }
 
+#[derive(Debug,Clone)]
 pub struct Fields {
-  id_block_size: usize,
-  id_cache_size: usize,
-  id_flush_size: usize,
-  quad_block_size: usize,
-  quad_cache_size: usize,
-  quad_flush_size: usize,
-  missing_flush_size: usize,
+  pub id_block_size: u64,
+  pub id_cache_size: usize,
+  pub id_flush_size: u64,
+  pub id_flush_top: usize,
+  pub id_flush_max_age: usize,
+  pub quad_block_size: u64,
+  pub quad_cache_size: usize,
+  pub quad_flush_size: u64,
+  pub quad_flush_top: usize,
+  pub quad_flush_max_age: usize,
+  pub missing_flush_size: u64,
 }
 
 impl Default for Fields {
@@ -70,10 +80,14 @@ impl Default for Fields {
     Self {
       id_block_size: 500_000,
       id_cache_size: 5_000,
-      id_flush_size: 5_000,
+      id_flush_size: 10_000_000,
+      id_flush_top: 200,
+      id_flush_max_age: 40,
       quad_block_size: 50_000,
-      quad_cache_size: 5_000,
-      quad_flush_size: 5_000,
+      quad_cache_size: 2_000,
+      quad_flush_size: 10_000_000,
+      quad_flush_top: 200,
+      quad_flush_max_age: 20,
       missing_flush_size: 2_000_000,
     }
   }
@@ -93,6 +107,8 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     if buf.is_empty() {
       let mut quad_updates = HashMap::new();
       quad_updates.insert(0, None);
+      let mut quad_count = HashMap::new();
+      quad_count.insert(0, 0);
       let xq = Self {
         storage: Mutex::new(storage),
         root: QTree::Quad {
@@ -102,8 +118,14 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         active_files: HashMap::new(),
         quad_cache: LruCache::new(fields.quad_cache_size),
         quad_updates: Arc::new(RwLock::new(quad_updates)),
+        quad_update_count: 0,
+        quad_update_age: HashMap::new(),
+        quad_count,
         id_cache: LruCache::new(fields.id_cache_size),
         id_updates: Arc::new(RwLock::new(HashMap::new())),
+        id_update_age: HashMap::new(),
+        id_update_count: 0,
+        id_count: HashMap::new(),
         missing_updates: HashMap::new(),
         missing_count: 0,
         next_quad_id: 1,
@@ -120,8 +142,14 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         active_files: HashMap::new(),
         quad_cache: LruCache::new(fields.quad_cache_size),
         quad_updates: Arc::new(RwLock::new(HashMap::new())),
+        quad_update_count: 0,
+        quad_update_age: HashMap::new(),
+        quad_count: meta.quad_count,
         id_cache: LruCache::new(fields.id_cache_size),
         id_updates: Arc::new(RwLock::new(HashMap::new())),
+        id_update_age: HashMap::new(),
+        id_update_count: 0,
+        id_count: HashMap::new(),
         missing_updates: HashMap::new(),
         missing_count: 0,
         next_quad_id: meta.next_quad_id,
@@ -154,23 +182,29 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
     quad_ids
   }
   pub async fn read_quad(&mut self, q_id: QuadId) -> Result<HashMap<RecordId,R>,Error> {
-    if let Some(Some(records)) = self.quad_updates.read().await.get(&q_id) {
-      return Ok(records.clone());
-    }
-    if let Some(records) = self.quad_cache.get(&q_id) {
-      return Ok(records.clone());
-    }
-    let qfile = quad_file(q_id);
-    let records = match self.open_file_r(&qfile).await? {
-      Some(mut s) => {
-        let mut buf = vec![];
-        s.read_to_end(&mut buf).await?;
-        R::unpack(&buf)?
-      },
-      None => HashMap::new(),
+    let mut records = {
+      if let Some(records) = self.quad_cache.get(&q_id) {
+        records.clone()
+      } else {
+        let qfile = quad_file(q_id);
+        let records = match self.open_file_r(&qfile).await? {
+          Some(mut s) => {
+            let mut buf = vec![];
+            s.read_to_end(&mut buf).await?;
+            R::unpack(&buf)?
+          },
+          None => HashMap::new(),
+        };
+        self.quad_cache.put(q_id, records.clone());
+        self.close_file(&qfile);
+        records
+      }
     };
-    self.quad_cache.put(q_id, records.clone());
-    self.close_file(&qfile);
+    if let Some(Some(xrecords)) = self.quad_updates.read().await.get(&q_id) {
+      for (id,r) in xrecords {
+        records.insert(*id,r.clone());
+      }
+    }
     Ok(records)
   }
   pub async fn read_quad_denorm(&mut self, q_id: QuadId) -> Result<Vec<(RecordId,R,Vec<R>)>,Error> {
@@ -210,31 +244,26 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
   }
   async fn insert_id(&mut self, id: RecordId, q_id: QuadId) -> Result<(),Error> {
     let b = self.id_block(id);
-    let ifile = self.id_file(id);
     if let Some(ids) = self.id_updates.write().await.get_mut(&b) {
       ids.insert(id, q_id);
-      return Ok(());
-    }
-    if let Some(mut ids) = self.id_cache.pop(&b) {
-      ids.insert(id, q_id);
-      self.id_updates.write().await.insert(b, ids);
+      self.id_update_count += 1;
+      if let Some(c) = self.id_count.get_mut(&b) {
+        *c += 1;
+      } else {
+        self.id_count.insert(b, 1);
+      }
       return Ok(());
     }
     {
-      let o_s = self.open_file_r(&ifile).await?;
-      if let Some(mut ids) = self.id_cache.pop(&b) {
-        ids.insert(id, q_id);
-        self.id_updates.write().await.insert(b, ids);
-        return Ok(());
-      }
-      let mut buf = vec![];
-      if let Some(mut s) = o_s {
-        s.read_to_end(&mut buf).await?;
-      }
-      let mut ids = unpack_ids(&buf)?;
+      let mut ids = HashMap::new();
       ids.insert(id, q_id);
       self.id_updates.write().await.insert(b, ids);
-      self.close_file(&ifile);
+      self.id_update_count += 1;
+      if let Some(c) = self.id_count.get_mut(&b) {
+        *c += 1;
+      } else {
+        self.id_count.insert(b, 1);
+      }
     }
     Ok(())
   }
@@ -247,31 +276,29 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         let record = records.get(*i).unwrap();
         self.insert_id(record.get_id(), *q_id).await?;
       }
-      let item_len = {
-        let mut item_len = None;
-        {
-          let mut qu = self.quad_updates.write().await;
-          if let Some(Some(items)) = qu.get_mut(&q_id) {
-            for i in ix {
-              let r = records.get(*i).unwrap();
-              items.insert(r.get_id(), r.clone());
-            }
-            item_len = Some(items.len());
+      {
+        let mut qu = self.quad_updates.write().await;
+        if let Some(Some(items)) = qu.get_mut(q_id) {
+          for i in ix {
+            let r = records.get(*i).unwrap();
+            items.insert(r.get_id(), r.clone());
           }
-        };
-        if item_len.is_none() {
-          let mut items = self.read_quad(*q_id).await?;
+        } else {
+          let mut items = HashMap::new();
           for i in ix {
             let r = records.get(*i).unwrap();
             items.insert(r.get_id(), r.clone());
           }
-          item_len = Some(items.len());
-          let mut qu = self.quad_updates.write().await;
           qu.insert(*q_id, Some(items));
         }
-        item_len.unwrap()
-      };
-      if item_len > self.fields.quad_block_size {
+        self.quad_update_count += ix.len() as u64;
+        if let Some(c) = self.quad_count.get_mut(q_id) {
+          *c += ix.len() as u64;
+        } else {
+          self.quad_count.insert(*q_id, ix.len() as u64);
+        }
+      }
+      if self.quad_count.get(q_id).cloned().unwrap_or(0) > self.fields.quad_block_size {
         self.split_quad(&q_id, &bbox).await?;
       }
     }
@@ -280,48 +307,172 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
   }
   pub async fn check_flush(&mut self) -> Result<(),Error> {
     // todo: parallel io
-    if self.quad_updates.read().await.len() >= self.fields.quad_flush_size {
-      self.quad_flush().await?;
+    if self.quad_update_count >= self.fields.quad_flush_size {
+      self.quad_flush_partial(self.fields.quad_flush_top).await?;
     }
-    if self.id_updates.read().await.len() >= self.fields.id_flush_size {
-      self.id_flush().await?;
+    if self.id_update_count >= self.fields.id_flush_size {
+      self.id_flush_partial(self.fields.id_flush_top).await?;
     }
-    if self.missing_updates.len() >= self.fields.missing_flush_size {
+    if self.missing_updates.len() as u64 >= self.fields.missing_flush_size {
       self.missing_flush().await?;
     }
     Ok(())
   }
-  pub async fn quad_flush(&mut self) -> Result<(),Error> {
+  pub async fn quad_flush_by_id(&mut self, q_id: QuadId) -> Result<(),Error> {
+    let o_rs = {
+      let mut qu = self.quad_updates.write().await;
+      qu.remove(&q_id).and_then(|v| v)
+    };
+    if o_rs.is_none() { return Ok(()) }
+    let rs: HashMap<u64,R> = o_rs.unwrap();
+    if rs.is_empty() { return Ok(()) }
+    let qfile = quad_file(q_id);
+    let mut s = self.open_file_rw(&qfile).await?;
+    let buf = {
+      if let Some(records) = self.quad_cache.get_mut(&q_id) {
+        for (r_id,r) in rs.iter() {
+          records.insert(*r_id, r.clone());
+        }
+        R::pack(&records)
+      } else {
+        let mut buf = vec![];
+        s.read_to_end(&mut buf).await?;
+        s.seek(io::SeekFrom::Start(0)).await?;
+        let mut records = R::unpack(&buf)?;
+        for (r_id,r) in rs.iter() {
+          records.insert(*r_id, r.clone());
+        }
+        self.quad_cache.put(q_id, records.clone());
+        R::pack(&records)
+      }
+    };
+    s.set_len(buf.len() as u64).await?;
+    s.write_all(&buf).await?;
+    s.flush().await?;
+    self.close_file(&qfile);
+    self.quad_update_count -= rs.len() as u64;
+    self.quad_count.insert(q_id, 0);
+    self.quad_update_age.insert(q_id, 0);
+    Ok(())
+  }
+  pub async fn quad_flush_all(&mut self) -> Result<(),Error> {
     // todo: parallel io
-    let qu = self.quad_updates.clone();
-    for (q_id,o_rs) in qu.write().await.drain() {
-      if o_rs.is_none() { continue }
-      let rs = o_rs.unwrap();
-      let qfile = quad_file(q_id);
-      let mut s = self.open_file_rw(&qfile).await?;
-      let buf = &R::pack(&rs);
-      s.set_len(buf.len() as u64).await?;
-      s.write_all(&buf).await?;
-      s.flush().await?;
-      self.quad_cache.put(q_id,rs);
-      self.close_file(&qfile);
+    let qs = self.quad_updates.read().await.keys().copied().collect::<Vec<QuadId>>();
+    for q_id in qs {
+      self.quad_flush_by_id(q_id).await?;
     }
     Ok(())
   }
-  pub async fn id_flush(&mut self) -> Result<(),Error> {
-    let iu = self.id_updates.clone();
-    for (b,ids) in iu.write().await.drain() {
-      let ifile = id_file_from_block(b);
-      let mut s = self.open_file_rw(&ifile).await?;
-      let buf = pack_ids(&ids);
-      s.set_len(buf.len() as u64).await?;
-      s.write_all(&buf).await?;
-      s.flush().await?;
-      self.id_cache.put(b,ids);
-      self.close_file(&ifile);
+  pub async fn quad_flush_partial(&mut self, n: usize) -> Result<(),Error> {
+    let mut qs = self.quad_updates.read().await.keys()
+      .filter(|q_id| { self.quad_count.get(q_id).copied().unwrap_or(0) > 0 })
+      .copied()
+      .collect::<Vec<QuadId>>();
+    qs.sort_unstable_by(|a,b| {
+      let ca = self.quad_count.get(a).copied().unwrap_or(0);
+      let cb = self.quad_count.get(b).copied().unwrap_or(0);
+      cb.cmp(&ca) // descending
+    });
+    for q_id in &qs[0..n.min(qs.len())] {
+      //println!["flush {} (top: {})", q_id, self.quad_count.get(q_id).copied().unwrap_or(0)];
+      self.quad_flush_by_id(*q_id).await?;
+    }
+    let mut queue = vec![];
+    for q_id in self.quad_updates.read().await.keys() {
+      let age = {
+        if let Some(age) = self.quad_update_age.get_mut(q_id) {
+          *age += 1;
+          age.clone()
+        } else {
+          self.quad_update_age.insert(*q_id, 1);
+          1
+        }
+      };
+      if age > self.fields.quad_flush_max_age {
+        //println!["flush {} (age)", q_id];
+        queue.push(*q_id);
+      }
+    }
+    for q_id in queue {
+      self.quad_flush_by_id(q_id).await?;
     }
     Ok(())
   }
+
+  pub async fn id_flush_by_block(&mut self, b: IdBlock) -> Result<(),Error> {
+    let o_ids = self.id_updates.write().await.remove(&b);
+    if o_ids.is_none() { return Ok(()) }
+    let ids = o_ids.unwrap();
+    let ifile = id_file_from_block(b);
+    let mut s = self.open_file_rw(&ifile).await?;
+    let buf = {
+      if let Some(xids) = self.id_cache.get_mut(&b) {
+        for (id,q_id) in ids.iter() {
+          xids.insert(*id, *q_id);
+        }
+        pack_ids(&xids)
+      } else {
+        let mut buf = vec![];
+        s.read_to_end(&mut buf).await?;
+        s.seek(io::SeekFrom::Start(0)).await?;
+        let mut xids = unpack_ids(&buf)?;
+        for (id,q_id) in ids.iter() {
+          xids.insert(*id,*q_id);
+        }
+        pack_ids(&xids)
+      }
+    };
+    s.set_len(buf.len() as u64).await?;
+    s.write_all(&buf).await?;
+    s.flush().await?;
+    self.id_update_count -= ids.len() as u64;
+    self.id_cache.put(b,ids);
+    self.id_count.insert(b, 0);
+    self.close_file(&ifile);
+    Ok(())
+  }
+  pub async fn id_flush_all(&mut self) -> Result<(),Error> {
+    let blocks = self.id_updates.read().await.keys().copied().collect::<Vec<IdBlock>>();
+    for b in blocks {
+      self.id_flush_by_block(b).await?;
+    }
+    Ok(())
+  }
+  pub async fn id_flush_partial(&mut self, n: usize) -> Result<(),Error> {
+    let mut qs = self.id_updates.read().await.keys()
+      .filter(|b| { self.id_count.get(b).copied().unwrap_or(0) > 0 })
+      .copied()
+      .collect::<Vec<QuadId>>();
+    qs.sort_unstable_by(|a,b| {
+      let ca = self.id_count.get(a).copied().unwrap_or(0);
+      let cb = self.id_count.get(b).copied().unwrap_or(0);
+      cb.cmp(&ca) // descending
+    });
+    for b in &qs[0..n.min(qs.len())] {
+      //println!["flush {} (top: {})", q_id, self.quad_count.get(q_id).copied().unwrap_or(0)];
+      self.id_flush_by_block(*b).await?;
+    }
+    let mut queue = vec![];
+    for b in self.id_updates.read().await.keys() {
+      let age = {
+        if let Some(age) = self.id_update_age.get_mut(b) {
+          *age += 1;
+          age.clone()
+        } else {
+          self.id_update_age.insert(*b, 1);
+          1
+        }
+      };
+      if age > self.fields.id_flush_max_age {
+        queue.push(*b);
+      }
+    }
+    for b in queue {
+      self.id_flush_by_block(b).await?;
+    }
+    Ok(())
+  }
+
   pub async fn missing_flush(&mut self) -> Result<(),Error> {
     if self.missing_updates.is_empty() { return Ok(()) }
     let m_id = self.next_missing_id;
@@ -338,8 +489,8 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
   }
   pub async fn flush(&mut self) -> Result<(),Error> {
     // todo: parallel io
-    self.quad_flush().await?;
-    self.id_flush().await?;
+    self.quad_flush_all().await?;
+    self.id_flush_all().await?;
     self.missing_flush().await?;
     Ok(())
   }
@@ -363,7 +514,9 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
       g.unwrap()
     };
     if let Some(records) = self.quad_updates.read().await.get(&q_id) {
-      return Ok(records.as_ref().and_then(|items| items.get(&id).cloned()));
+      if let Some(r) = records.as_ref().and_then(|items| items.get(&id).cloned()) {
+        return Ok(Some(r));
+      }
     }
     if let Some(records) = self.quad_cache.get(&q_id) {
       return Ok(records.get(&id).cloned());
@@ -396,28 +549,15 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
   }
   pub async fn split_quad(&mut self, q_id: &QuadId, bbox: &BBox) -> Result<(),Error> {
     let qfile = quad_file(*q_id);
-    let records = {
-      let qu = self.quad_updates.write().await.remove(q_id);
-      let qc = self.quad_cache.pop(q_id);
-      match (qu,qc) {
-        (Some(Some(rs)),_) => {
-          self.lock_file(&qfile).await;
-          rs
-        },
-        (Some(None),_) => panic!["tried to split an empty quad"],
-        (_,Some(rs)) => {
-          self.lock_file(&qfile).await;
-          rs
-        },
-        (None,None) => {
-          let mut buf = vec![];
-          if let Some(mut s) = self.open_file_r(&qfile).await? {
-            s.read_to_end(&mut buf).await?;
-          }
-          R::unpack(&buf)?
-        },
+    let records = self.read_quad(*q_id).await?;
+    {
+      let mut st = self.storage.lock().await;
+      if st.exists(&qfile).await {
+        st.remove(&qfile).await?;
       }
-    };
+    }
+    self.lock_file(&qfile).await;
+    self.quad_count.insert(*q_id, 0);
     let (nx,ny) = (2,2);
     let mut quads = vec![];
     for i in 0..nx {
@@ -439,9 +579,10 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         o_p = self.get_position(&r).await?;
       }
       if let Some(p) = o_p {
-        let i = quads.iter().position(|(b,_)| overlap(&p, &b)).unwrap();
-        let q = quads.get_mut(i).unwrap();
-        q.1.insert(*r_id,r.clone());
+        if let Some(i) = quads.iter().position(|(b,_)| overlap(&p, &b)) {
+          let q = quads.get_mut(i).unwrap();
+          q.1.insert(*r_id,r.clone());
+        }
       } else {
         panic!["missing record in split quad"];
         //self.missing_updates.insert(r.get_id(),r);
@@ -457,7 +598,9 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         }
         if q.1.is_empty() {
           self.quad_updates.write().await.insert(*q_id, None);
+          self.quad_count.insert(*q_id, 0);
         } else {
+          self.quad_count.insert(*q_id, q.1.len() as u64);
           self.quad_updates.write().await.insert(*q_id, Some(q.1));
         }
         nchildren.push(QTree::Quad { id: *q_id, bbox: q.0.clone() });
@@ -469,7 +612,9 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
         }
         if q.1.is_empty() {
           self.quad_updates.write().await.insert(id, None);
+          self.quad_count.insert(id, 0);
         } else {
+          self.quad_count.insert(id, q.1.len() as u64);
           self.quad_updates.write().await.insert(id, Some(q.1));
         }
         nchildren.push(QTree::Quad { id, bbox: q.0.clone() });
@@ -639,6 +784,7 @@ impl<S,R> XQ<S,R> where S: RW, R: Record {
       id_block_size: self.fields.id_block_size,
       quad_block_size: self.fields.quad_block_size,
       root: self.root.clone(),
+      quad_count: self.quad_count.clone(),
     }
   }
   pub async fn save_meta(&mut self) -> Result<(),Error> {
